// Imagine++ project
// Project:  Shocks
// Author:   keriven
// Date:     2007/10/18

#include <Imagine/Graphics.h>
using namespace Imagine;
#include <cstdlib>
#include <ctime>
using namespace std;

///////////////////////////////
// Constants of program
const int width=256;    // Window width
const int height=256;   // Window height
const int ball_size=4;  // Radius of ball
const int nb_balls=30;  // Number of balls

///////////////////////////////
// Random generator
// Call it only once, before Random()
void initRand() {
    srand((unsigned int)time(0));
}

// Between a and b
int random(int a,int b) {
    return a+(rand()%(b-a+1));
}

///////////////////////////////
// Random position and speed
void initBall(double &x,double &y,double &u,double &v,Color &c){
    x=random(ball_size,width-ball_size);
    y=random(ball_size,height-ball_size);
    u=random(0,4);
    v=random(0,4);
    c=Color(byte(random(0,255)),
            byte(random(0,255)),
            byte(random(0,255)));
}

///////////////////////////////
// Display of ball
void displayBall(double x,double y,Color col) {
    fillRect(int(x)-ball_size,int(y)-ball_size,
             2*ball_size+1,2*ball_size+1,col);
}

///////////////////////////////
// Elastic shock of two spherical balls
// cf labo.ntic.org
#include <cmath>
void shockBalls(double&x1,double&y1,double&u1,double&v1,
                double&x2,double&y2,double&u2,double&v2)
{
    // Distance
    double o2o1x=x1-x2,o2o1y=y1-y2;
    double d=sqrt(o2o1x*o2o1x+o2o1y*o2o1y);
    if(d==0) return; // Same center?
    // Frame (o2,x,y)
    double Vx=u1-u2,Vy=v1-v2;
    double V=sqrt(Vx*Vx+Vy*Vy);
    if(V==0) return; // Same speed?
    // Next frame V (o2,i,j)
    double ix=Vx/V,iy=Vy/V,jx=-iy,jy=ix;
    // Height
    double H=o2o1x*jx+o2o1y*jy;
    // Angle
    double th=acos(H/d),c=cos(th),s=sin(th);
    // Speed after shock in (o2,i,j)
    double v1i=V*c*c,v1j=V*c*s,v2i=V*s*s,v2j=-v1j;
    // In original frame (O,x,y)
    u1=v1i*ix+v1j*jx+u2;
    v1=v1i*iy+v1j*jy+v2;
    u2+=v2i*ix+v2j*jx;
    v2+=v2i*iy+v2j*jy;
}

///////////////////////////////
// Motion of ball
void moveBall(double x[],double y[],double u[],double v[],int i)
{   // Bounce on left and right borders
    if(x[i]+u[i]>width-ball_size || x[i]+u[i]<ball_size)
        u[i]=-u[i];
    // Bound on top and bottom borders
    if(y[i]+v[i]<ball_size || y[i]+v[i]>height-ball_size)
        v[i]=-v[i];
    for(int j=0; j<nb_balls; j++) {
        if(j==i)
            continue;
        if(abs(x[i]+u[i]-x[j])<2*ball_size &&
           abs(y[i]+v[i]-y[j])<2*ball_size) {
            shockBalls(x[i],y[i],u[i],v[i],x[j],y[j],u[j],v[j]);
        }
    }
    // Update position
    x[i]+=u[i];
    y[i]+=v[i];
}

///////////////////////////////
// Main function
int main() {
    // Opening window
    openWindow(width,height);
    // Position and speed of balls
    double xb[nb_balls],yb[nb_balls],ub[nb_balls],vb[nb_balls];    
    Color cb[nb_balls]; // Colors of balls
    initRand();
    for(int i=0; i<nb_balls; i++) {
        initBall(xb[i],yb[i],ub[i],vb[i],cb[i]);
        displayBall(xb[i],yb[i],cb[i]);
    }
    // Main loop
    while(true) {
        milliSleep(25);
        noRefreshBegin();
        for(int i=0; i<nb_balls; i++) {
            displayBall(xb[i],yb[i],WHITE);
            moveBall(xb,yb,ub,vb,i);
            displayBall(xb[i],yb[i],cb[i]);
        }
        noRefreshEnd();
    }
    endGraphics();
    return 0;
}
